// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "PokerGame.hpp"

using namespace Dawn;

std::shared_ptr<PokerPlayer> PokerGame::addNewPlayer() {
  auto player = std::make_shared<PokerPlayer>(
    weak_from_this(), this->players.size()
  );
  assertFalse(this->players.size() == PLAYER_COUNT_MAX, "Too many players.");
  this->players.push_back(player);
  return player;
}

std::shared_ptr<PokerPlayer> PokerGame::getCurrentBetter() {
  auto nextIndex = this->getNextBetterIndex();
  if(nextIndex == 0xFF) return nullptr;
  return this->players[nextIndex];
}

void PokerGame::newGame() {
  assertTrue(this->players.size() >= PLAYER_COUNT_MIN, "Not enough players.");
  this->newRound();
  this->smallBlind = POKER_BLIND_SMALL_DEFAULT;
  this->bigBlind = POKER_BLIND_BIG_DEFAULT;

  auto it = this->players.begin();
  while(it != this->players.end()) {
    auto player = *it;
    player->setChips(POKER_PLAYER_CHIPS_DEFAULT);
    player->isOut = false;
    ++it;
  }
  
  this->setDealer(0x00);
}

void PokerGame::newRound() {
  this->deck.clear();
  Card::fillDeck(this->deck);

  this->grave.clear();
  this->community.clear();
  this->pots.clear();
  this->pots.push_back(PokerPot());

  this->hasFlopped = false;
  this->hasTurned = false;
  this->hasRivered = false;

  auto it = this->players.begin();
  while(it != this->players.end()) {
    auto player = *it;
    player->hand.clear();
    player->currentBet = 0;
    player->isFolded = false;
    player->isShowingHand = false;
    player->hasBetThisRound = false;
    player->timesRaised = 0;
    player->currentBet = 0;
    ++it;
  }

  this->setDealer(this->dealerIndex + 0x01);
}

void PokerGame::newBettingRound() {
  auto it = this->players.begin();
  while(it != this->players.end()) {
    auto player = *it;
    player->hasBetThisRound = false;
    player->timesRaised = 0;
    ++it;
  }

  this->betterIndex = this->bigBlindIndex;
  this->betterIndex = this->getNextBetterIndex();
}

uint8_t PokerGame::getNextBetterIndex() {
  uint8_t j, i;
  for(i = 0; i < this->players.size(); i++) {
    j = (i + this->betterIndex) % this->players.size();
    auto player = this->players[j];
    if(player->needsToBetThisRound()) return j;
  }
  return 0xFF;
}

void PokerGame::takeBlinds() {
  auto playerSmallBlind = this->players[this->smallBlindIndex];
  auto playerBigBlind = this->players[this->bigBlindIndex];

  playerSmallBlind->bet(this->smallBlind);
  playerBigBlind->bet(this->bigBlind);

  playerSmallBlind->hasBetThisRound = false;
  playerBigBlind->hasBetThisRound = false;
}

void PokerGame::setDealer(const uint8_t dealer) {
  uint8_t i, k;
  std::shared_ptr<PokerPlayer> player;
  bool_t foundDealer;
  bool_t foundSmall;

  foundDealer = false;
  foundSmall = false;
  this->dealerIndex = dealer;

  for(i = 0; i < this->players.size(); i++) {
    k = (dealer + i) % this->players.size();
    player = this->players[k];
    if(player->isOut) continue;
    if(!foundDealer) {
      this->dealerIndex = k;
      foundDealer = true;
    } else if(!foundSmall) {
      this->smallBlindIndex = k;
      foundSmall = true;
    } else {
      this->bigBlindIndex = k;
      break;
    }
  }
}

uint8_t PokerGame::getRemainingBettersCount() {
  uint8_t count = 0;
  auto it = this->players.begin();
  while(it != this->players.end()) {
    if((*it)->needsToBetThisRound()) count++;
    ++it;
  }
  return count;
}

uint8_t PokerGame::getRemainingPlayersCount() {
  uint8_t count = 0;
  auto it = this->players.begin();
  while(it != this->players.end()) {
    if(!(*it)->isFolded && !(*it)->isOut) count++;
    ++it;
  }
  return count;
}

int32_t PokerGame::getCurrentCallValue() {
  assertTrue(this->pots.size() > 0, "No pots?");
  return this->pots.back().call;
}

void PokerGame::burnCard() {
  assertTrue(this->deck.size() > 0, "No cards to burn.");
  auto card = this->deck.back();
  this->deck.pop_back();
  this->grave.push_back(card);
}

void PokerGame::dealCard(PokerPlayer &player) {
  assertTrue(this->deck.size() > 0, "No cards to deal.");
  auto card = this->deck.back();
  this->deck.pop_back();
  player.hand.push_back(card);
}

void PokerGame::dealToEveryone(const uint8_t count) {
  for(uint8_t i = 0; i < count; i++) {
    auto it = this->players.begin();
    while(it != this->players.end()) {
      this->dealCard(*(*it));
      ++it;
    }
  }
}

void PokerGame::turn(const uint8_t count) {
  uint8_t c = count;
  if(c == 0xFF) c = this->getCountOfCardsToTurn();
  assertTrue(this->deck.size() >= c, "Not enough cards to turn.");
  for(uint8_t i = 0; i < c; i++) {
    auto card = this->deck.back();
    this->deck.pop_back();
    this->community.push_back(card);
  }
}

uint8_t PokerGame::getCountOfCardsToTurn() {
  if(!this->hasFlopped) return 3;
  if(!this->hasTurned) return 1;
  if(!this->hasRivered) return 1;
  assertUnreachable("No more cards to turn.");
}